"""Interface to read Item Response Data stored in a .csv file.

*** MAIN CLASS:
Table --- iterable read-only interface to a csv file,
    accessing fields by their names as defined in the first row of the file.

NOTE: This Table class assumes csv dialect = 'excel'.
The first row in the file MUST include strings corresponding to the Table fields property.

Among all possible csv format parameters,
the Table allows only delimiter and quotechar to be assigned by the user.
csv.quoting is always csv.QUOTE_NONNUMERIC, i.e. all numeric fields are read as float values.
"""
import csv
from . import table_reader


# -----------------------------------------------------------------
class Table(table_reader.Table):
    """Reader for tabular data stored in CSV file,
    using python standard csv library
    """
    def __init__(self, file_path, fields=None, delimiter=',', quotechar='"', **kwargs):
        super().__init__(file_path, fields, **kwargs)
        self.delimiter = delimiter
        self.quotechar = quotechar

    def __iter__(self):
        """Generator of file rows for read-only access.
        :return: iterator object over file rows, except header
        """
        with self.table_id.open('r', newline='') as f:
            reader = csv.reader(f, delimiter=self.delimiter,
                                quotechar=self.quotechar,
                                quoting=csv.QUOTE_NONNUMERIC)
            if self.fields is None:
                field_index = None
            else:
                header = next(reader)
                # = first row
                field_index = [header.index(f_i) for f_i in self.fields]
            for (row_n, row) in enumerate(reader):
                if row_n % self.sample_factor == 0:
                    if field_index is not None:
                        row =[row[f_i] for f_i in field_index]
                    # else keep all fields
                    row = self.process_row(row)
                    if row is not None:
                        yield row

    def preview(self, max_records=10):
        """Generator of rows for read-only access.
        :return: iterator object over file rows, except header
        """
        with self.table_id.open('r', newline='') as f:
            reader = csv.reader(f, delimiter=self.delimiter, quotechar=self.quotechar)
            for (row_n, row) in enumerate(reader):
                if row_n > max_records:
                    break
                yield row
        # return result


# ------------------------------------------------------------ TEST:
# if __name__ == '__main__':
#     from pathlib import Path
#
#     HAQ_PATH = Path.home() / 'Documents/LeijonKTH/heartech/HA-QualityRegistry'
#     # top-dir for everything to be used here
#
#     ioiha_data_path = HAQ_PATH / 'IRT-IOI-HA' / 'IOI-HA-data'
#     # to ioi-ha data sets OTHER THAN the Swedish Quality Registry
#
#     ioiha_kramer = ioiha_data_path / 'Kramer'
#
#     kramer_file = ioiha_kramer / 'moeder_final - kopieforArne.sav'
#     kramer_file = ioiha_data_path / 'Kramer' / 'NLSH-Kramer-T2IOI-HA.sav'
#     kramer_file = kramer_file.with_suffix('.csv')
#
#     print('Testing kramer_file preview:')
#     reader = Table(kramer_file).preview(10)
#     row0 = next(reader)
#     print('header= ', row0)
#     for (n, line) in enumerate(reader):
#         print(line)
#
#     # spss = Table(kramer_file)
#     # print('spss.header = ', spss.header)
#     #
#     # data = spss.preview(10)
#     # print('spss.preview(10) =\n', data)
#
#     # -------------------------------- test with missing-values recoding:
#     print('Testing kramer_file Table: ')
#     csvt = Table(kramer_file,
#                  # fields=['ioi_v1', 'ioi_v2', 'ioi_v3'],
#                  fields=[f'T2IOIHA{i+1}' for i in range(7)],
#                  missing_values=[9999999.0],
#                  field_types=str)
#     print('repr(csvt)= ', csvt)
#
#     for (n, row) in enumerate(csvt, start=1):
#         if n > 10:
#             break
#         print(f'row #{n} = {row}')
#
#     # -------------------------------------------------------------------
#     print('\n*** final test with missing_values, type_cast, recode_fcn ***\n')
#
#     def recode_missing(row):
#         for i in range(len(row)):
#             if row[i] is None:
#                 row[i] = 0
#
#     def recode_fields14(row):
#         """IOI-HA items 1,2,3,4 coded as 0-4,
#         recode to 1-5,
#         BEFORE recoding missing-data as zero.
#         """
#         for i in [1, 4]:
#             if row[i-1] is not None:
#                 row[i-1] += 1
#     # --------------------------------------------
#
#     csvt = Table(kramer_file,
#                  # fields=['ioi_v1', 'ioi_v2', 'ioi_v3'],
#                  fields=[f'T2IOIHA{i+1}' for i in range(7)],
#                  missing_values=[9999999.0],
#                  field_types=int,
#                  recode_fcns=[recode_fields14, recode_missing])
#
#     print('repr(csvt)= ', csvt)
#     for (n, row) in enumerate(csvt, start=1):
#         # if any(r_i is not None and (r_i <= 0 or r_i >= 5) for r_i in row):
#         if any(r_i is not None and (r_i > 5) for r_i in row):
#             print(f'row #{n} = {row}')
#         if n < 10:
#             print(f'row#{n} = {row}')
#
#     print(f'len(csvt)={len(csvt)}')
#     n_all_zero = sum(all(r_i == 0 for r_i in r)
#                      for r in csvt)
#     print(f'Number of all-zero records = {n_all_zero}')
#



"""Package ItemResponseCalc implements Bayesian analysis of
subjective responses to items in a questionnaire,
using Item Response Theory (IRT) applied with
the Graded Response Model (Samejima, 1997; Fox, 2010).

This package is only intended for the analysis of data with ORDINAL responses,
e.g., the IOI-HA questionnaire for users' satisfaction with their hearing aids.

The IRT Graded Response Model treats subjects' responses
as determined by the outcome of a latent individual trait variable,
i.e., similar to the latent internal "sensation"
variable assumed to determine responses in psycho-physical experiments.
This package uses a logistic distribution for the latent variables.

Each questionnaire item corresponds to exactly ONE individual trait variable,
but each trait variable may determine the subject's responses to several items.

Therefore, for a questionnaire with I separate items,
the model can learn D <= I trait variables, based on the observed responses.
The model learning procedure automatically determines how many
different trait variables are needed to 'explain' the recorded data.

NOTE: In most modules of this package,
the ordinal response to the i-th item is encoded as
an integer in {0, 1,..., L_i-1},
where L_i = the number of ordinal response levels for the i-th item.
Any value outside this range is interpreted as a missing response.

However, in the input data files,
the recorded responses are usually encoded as integers in {1, 2,..., L_i},
with missing values encoded as zero.
This notation is used in module item_response_data.
and in some results presentations where response levels are identified.


*** Usage

Copy the template script `run_irt.py` to your working directory, rename it, and
edit the copy as suggested in the template, to specify
    - your experimental layout,
    - your input data file(s),
    - a directory where all output result files will be stored.

*** Main Modules

item_response_data:
    Interface to one or several data sets containing individual responses
    from respondents recruited from one or several different population(s).
    See this module for information about input files and data formats.

item_response_model:
    The core of the Bayesian IRT model adapted to all observed data.

item_scale:
    Classes representing the item-specific model parameters,
    defining the connection between traits to ordinal responses, for each item.

item_respondents:
    Classes representing individual- and population-specific models
    defining trait distributions for each individual and each population.

ir_display:
    Functions and classes to display all analysis results as figures and tables.
    
table_reader:
    Defines class Table for input data in several accepted file formats.
    The present version can read tabulated data from sources of type
    xlsx, csv, sql, sqlite3.

Analysis results are saved as figures and tables in the selected result directory.
After running an analysis, the logging output briefly explains
the analysis results presented in figures and tables.

*** References
Arne Leijon and Harvey Dillon and Louise Hickson
and Martin Kinkel and Sophia E Kramer and Peter Nordquist (2020):
    Analysis of Data from the International Outcome Inventory for Hearing Aids (IOI-HA)
    using Bayesian Item Response Theory.
    Manuscript in preparation.

J.-P. Fox (2010). *Bayesian Item Response Modeling: Theory and Applications*.
    Statistics for Social and Behavioral Sciences. Springer.
G. N. Masters (1982). A Rasch model for partial credit scoring.
    *Psychometrika*, 47(2):149–174.
F. Samejima (1997). Graded response model.
    In W. J. v. D. Linden and R. K. Hambleton, eds.,
    *Handbook of Modern Item Response Theory*, p. 85–100. Springer, New York.


*** Version History

2019-08-24, first functional version with automatic selection of trait dimensionality.
2020-06-08, first version uploaded to PyPi, tested with several IOI-HA data sets.
"""

__name__ = 'ItemResponseCalc'
__version__ = '0.5.0'
__all__ = ['__version__', 'run_irt', 'show_irt']





"""Functions and classes to display results from an OrdinalItemResponseModel object.
"""
# ***** check special chars for latex and txt variants

import numpy as np
from itertools import cycle, product

import matplotlib.pyplot as plt
import logging

logger = logging.getLogger(__name__)

# --------------------------- Default format parameters:
FMT = {'table_format': 'latex',  # or 'tab' for tab-delimited tables
       'figure_format': 'pdf',   # or 'eps', or 'png', or 'pdf', for saved plots
       'colors': 'rbg',  # color cycle to separate results in plots
       'line_styles': ['solid', 'dashed', 'dashdot', 'dotted'],
       'markers': 'o^sDv',  # plot symbols, each repeated with each fillstyle
       'fillstyle': ['full', 'none'], # marker style
       'credibility': 'Credibility',  # heading in tables
       'correlation': 'Correlation',  # heading in tables
       }
# = module-global dict with default settings for display details
# that may be changed by user
# NOTE: Largest number of combinations of color, line_style, marker, fillstyle
# is obtained if the number of alternatives are NOT divisible by each other.

TABLE_FILE_SUFFIX = {'latex': '.tex', 'tab': '.txt'}
# = mapping table_format -> file suffix

# ------------------------- Matplotlib settings:
plt.rcParams.update({'figure.max_open_warning': 0})
# suppress warning for many open figures
pad_inches = plt.rcParams['savefig.pad_inches']
plt.rcParams['savefig.pad_inches'] = 0.
plt.rcParams['savefig.bbox'] = 'tight'
# *** allow user to set other plt.rcParams ? ***
# -----------------------------------------------


def set_format_param(**kwargs):
    """Set / modify module-global format parameters
    :param kwargs: dict with format variables
        to replace the default values in FMT
    :return: None
    """
    # *** allow user to set other plt.rcParams here ? ***
    for (k, v) in kwargs.items():
        k = k.lower()
        if k not in FMT:
            logger.warning(f'Format setting {k}={repr(v)} is not known, not used')
        FMT[k] = v


def _percent():
    """Percent sign for tables
    :return: str
    """
    return '\\%' if FMT['table_format'] == 'latex' else '%'


# ---------------------------- Main Result Classes
class FigureRef:
    """Reference to a single graph instance
    """
    def __init__(self, ax, name=None, path=None):
        """
        :param ax: Axes instance containing the graph
        :param path: Path to directory where figure has been saved
        :param name: (optional) updated name of figure file
        """
        self.ax = ax
        self.name = name
        self.path = path

    def __repr__(self):
        return (f'FigureRef(ax= {repr(self.ax)}, ' +
                f'name= {repr(self.name)}), path= {repr(self.path)}')

    @property
    def fig(self):
        return self.ax.figure

    def save(self, path, name=None):
        """Save figure to given path
        :param path: Path to directory where figure is saved
        :param name: (optional) updated name of figure file
        :return: None
        Result: updated properties path, name
        """
        if name is None:
            name = self.name
        name = _clean_file_name(name)
        path.mkdir(parents=True, exist_ok=True)
        f = (path / name).with_suffix('.' + FMT['figure_format'])
        self.fig.savefig(str(f))
        self.path = path
        self.name = f.name

    def mapped_y_axis(self, y, y_mapped, y_label='', fontsize=14):
        """Add a second y_axis to self.ax with ticks placed at
        y values corresponding to uniform steps in y_mapped values
        but labelled by the y_mapped values
        :param y: long sequence of y_values along original y axis
        :param y_mapped: corresponding transformed y_values for new y_ticks
            len(y_mapped) == len(y)
            y and y_mapped monotonically increasing
        :param y_label: label of second y-axis
        :param fontsize: (optional) fontsize of y_label
        :return: None
        """
        ax2 = self.ax.twinx()
        ax2.set_ylim(self.ax.get_ylim())
        ymap_ticks = _nice_ticks(np.amin(y_mapped), np.amax(y_mapped))
        # = uniform ticks in y_mapped scale
        ymap_ticklabels = [f'{tick:.1f}' for tick in ymap_ticks]
        y_ticks = np.interp(ymap_ticks, y_mapped, y)
        # = transformed to corresponding y scale positions
        ax2.set_yticks(y_ticks)
        ax2.set_yticklabels(ymap_ticklabels)
        ax2.set_ylabel(y_label, fontsize=fontsize)
        self.fig.tight_layout()


class TableRef:
    """Reference to a single table instance,
    formatted in LaTeX OR plain tabulated txt versions
    """
    def __init__(self, text=None, name=None, path=None):
        """
        :param text: single string with all table text incl. newline, tab chars.
        :param path: (optional) Path to directory where tables are saved
        :param name: (optional) updated file name, with or without suffix
            suffix is determined by FMT['table_format'] anyway
        """
        # store table parts instead *****???
        self.text = text
        self.name = name
        self.path = path

    def __repr__(self):
        return (f'TableRef(text= text, ' +    # fmt= {repr(self.fmt)}, ' +
                f'name= {repr(self.name)}), path= {repr(self.path)}')

    def save(self, path, name=None):
        """Save table to file.
        :param path: Path to directory where tables are saved
        :param name: (optional) updated file name, with or without suffix
            suffix is determined by FMT['table_format'] anyway
        :return: None
        Result: updated properties path, name
        """
        if name is None:
            name = self.name
        name = _clean_file_name(name)
        path.mkdir(parents=True, exist_ok=True)   # just in case
        f = (path / name).with_suffix(TABLE_FILE_SUFFIX[FMT['table_format']])
        if self.text is not None and len(self.text) > 0:
            f.write_text(self.text, encoding='utf-8')
        self.path = path
        self.name = f.name


# ------------------------------------------------ plot routines

def fig_log_likelihood(learned_ll, n_users,
                       title='',
                       # label=None,
                       fontsize=14,
                       name=None,
                       **kwargs):
    """plot VI learning result
    :param learned_ll = list log-likelihood values from a learned ItemResponseModel
    :param n_users = scalar number of users included in the total log-likelihood
    :param title = (optional) figure title
    :param name = (optional) figure name for file
    :param fontsize = (optional) font size for axis labels
    :param kwargs = (optional) dict with any additional arguments for plot function
    :return: None
    """
    f_logprob, ax = plt.subplots()
    t1 = 1
    # for ll in learned_ll:
    t = [t1 + n for n in range(len(learned_ll))]
    # n_train = ham.n_training_users
    ax.plot(t, np.array(learned_ll) / n_users, **kwargs)
    # t1 = t[-1] + 1
    ax.set_xlim(0, t[-1]+1)
    ax.set_xlabel('Learning Iterations', fontsize=fontsize)
    ax.set_ylabel('Log-likelihood / N', fontsize=fontsize)
    if 0 < len(title):
        ax.set_title(title)
    return FigureRef(ax, name=name)


def fig_response_freq(count, name=None, fontsize=14):
    """Generate a plot with relative frequencies of response counts for all items
    :param count: list of Counter objects  *** dict ?
        count[i][0] = number of missing responses for i-th item
        count[i][l] = number of responses in l-th ordinal level for i-th item
    :param fontsize = (optional) font size for axis labels
    :return: figure object *** FigureRef object ???
    """
    fig, ax = plt.subplots()
    for (i, (c, col, (m, fill_style))) in enumerate(zip(count,  # total_count,
                                                        cycle(FMT['colors']),
                                                        cycle(product(FMT['markers'],
                                                                      FMT['fillstyle'])))):
        x_i = np.arange(1 + max(*c.keys()))  # key = 0 means missing data
        y_i = np.array([c[r] for r in x_i])
        y_i = y_i / np.sum(y_i)
        y_i *= 100  # in percent
        ax.plot(x_i, y_i, label = f'Q{i+1}', color=col,
                marker=m, fillstyle=fill_style)
    ax.set_xticks(np.arange(0, 6))
    ax.set_xticklabels(['Missing'] + [f'R={i+1}' for i in range(5)])
    ax.set_xlabel('Item Response', fontsize=fontsize)
    ax.set_ylabel('Rel. Frequency (%)', fontsize=fontsize)
    ax.legend(loc='best')
    return FigureRef(ax, name=name)


def fig_response_prob(t, p,
                      tau=None,
                      name=None,
                      x_label='',
                      y_label='',
                      fontsize=14
                      ):
    """figure with response prob vs. trait
    :param t: 1D array with trait values
    :param p: 2D array with response probabilities
        p[n, l] = P{l-th response | t[n]}
    :param tau: (optional) 1D array with trait thresholds
    :param name: (optional) string for plot file name
    :param x_label: string for x-axis label
    :param y_label: string for y-axis label
    :param fontsize: (optional) fontsize for x_ and y_label
    :return: FigureRef object
    """
    fig, ax = plt.subplots()
    for (l, (p_l, c, ls)) in enumerate(zip(p.T,
                                           cycle(FMT['colors']),
                                           cycle(FMT['line_styles']))):
        ax.plot(t, p_l, color=c, linestyle=ls, label=f'R={1+l}')
    if tau is not None:
        # plot ticks at thresholds
        x = [tau, tau]
        y = [np.zeros(len(tau)), 0.1 * np.ones(len(tau))]
        ax.plot(x, y, color='k', linewidth=1., linestyle='solid')
    ax.set_ylim(0., 1.)
    ax.set_xlabel(x_label, fontsize=fontsize)
    ax.set_ylabel(y_label, fontsize=fontsize)
    ax.legend(loc='best')
    return FigureRef(ax, name=name)


def fig_percentiles(perc,
                    y_label,
                    x_ticklabels,
                    case_labels=None,
                    x_label=None,
                    x_offset=0.1,
                    x_space=0.5,
                    fontsize=14,
                    name=None,
                    **kwargs):
    """create a figure with trait percentile results
    :param perc: 3D (or 2D) array with trait percentiles
        perc[c, p, t] = p-th percentile of t-th variable for c-th case
        percentiles plotted vertically vs t-variables horizontally,
    :param y_label: string for y axis label
    :param x_label: string for x-axis label
    :param x_ticklabels: list of strings with labels for x_ticks,
        one for each value in rows perc[..., :]
        len(x_ticklabels) == perc.shape[-1] == number of traits
    :param case_labels: list of strings with labels for cases, if more than one
        len(case_labels) == perc.shape[0] if perc.ndim == 3
    :param x_offset: (optional) horizontal space between case plots for each x_tick
    :param x_space: (optional) min space outside min and max x_tick values
    :param fontsize: (optional) fontsize for x and theta labels
    :param kwargs: (optional) dict with any additional keyword arguments for plot commands.
    :return: None  # fig object with single plot axis with all results
    """
    fig, ax = plt.subplots()
    if perc.ndim == 2:
        perc = perc[np.newaxis,...]
        case_labels = [None]
    (n_cases, n_perc, n_xticks) = perc.shape
    x = np.arange(0., n_xticks) - x_offset * (n_cases - 1) / 2
    for (c_key, c_y, c, (m, fs)) in zip(case_labels, perc,
                                      cycle(FMT['colors']),
                                      cycle(product(FMT['markers'],
                                                    FMT['fillstyle']))):
        ax.plot(np.tile(x, (2, 1)), c_y[[0, 2], :],
                linestyle='solid', color=c, **kwargs)
        ax.plot(x, c_y[1, :], linestyle='', color=c,
                marker=m, fillstyle=fs,  # markeredgecolor=c,  # markerfacecolor='w',
                label=c_key, **kwargs)
        x += x_offset
    (x_min, x_max) = ax.get_xlim()
    x_min = min(x_min, -x_space)
    x_max = max(x_max, n_xticks - 1 + x_space)
    ax.set_xlim(x_min, x_max)
    ax.set_xticks(np.arange(n_xticks))
    ax.set_xticklabels(x_ticklabels)
    ax.set_ylabel(y_label, fontsize=fontsize)  # + ' (' + y_unit + ')')
    ax.set_xlabel(x_label, fontsize=fontsize)
    if n_cases > 1:
        # make space for legend
        (x_min, x_max) = ax.get_xlim()
        ax.set_xlim(x_min, x_max + 0.6)
        ax.legend(loc='best')
    if name is None:
        name = _clean_file_name(y_label)
    return FigureRef(ax, name=name)


def fig_credible_diff(c_diff,
                      x_labels,
                      x_label=None,
                      title=None,
                      cred_levels=None,
                      fontsize=10
                      ):
    """Nice color plot to show jointly credible differences
    :param c_diff: list of tuples ((i, j), p), indicating that
        x_labels[i] > x-labels[j] with joint credibility p
    :param x_labels: list of labels of compared categories
    :param x_label: (optional) axis label
    :param title: (optional) string with plot title
    :param cred_levels: (optional) list of joint-credibility values, in DECREASING order
    :param fontsize: (optional) fontsize for x_labels in plot
    :return:
    """
    if cred_levels is None:
        cred_levels = [.99, .95, .9, .8, .7]
    marker_sizes = [15, 12, 9, 6, 3]

    cred_levels = np.asarray(cred_levels)
    symbol_size = None  # use previous size to detect change

    # ------------------------------------------------
    def select_symbol(p, prev_size):
        """Determine symbol size and legend text
        """
        i_size = np.nonzero(cred_levels < p)[0][0]
        size = marker_sizes[i_size]
        if size != prev_size:
            return size, f'> {cred_levels[i_size]:.0%}'
        else:
            return size, None
    # ------------------------------------------------

    fig, ax = plt.subplots()
    for ((i, j), p) in c_diff:
        symbol_size, label = select_symbol(p, symbol_size)
        if symbol_size > 0:
            s, = ax.plot(i, len(x_labels) - 1 - j, 'sr', markersize=symbol_size)
            if label is not None:
                s.set_label(label)
    ax.set_xlim([-0.5, len(x_labels) - 0.5])
    ax.set_ylim([-0.5, len(x_labels) - 0.5])
    ax.set_xticks(np.arange(len(x_labels)))
    ax.set_yticks(np.arange(len(x_labels)))
    ax.set_xticklabels(x_labels,
                       fontsize=fontsize,
                       **_x_tick_style(x_labels) )
    ax.set_yticklabels(x_labels[::-1],  # reversed order
                       fontsize=fontsize,
                       rotation='horizontal',
                       horizontalalignment='right'
                       )
    if x_label is not None:
        ax.set_xlabel(x_label, fontsize=fontsize + 4)
        ax.set_ylabel(x_label, fontsize=fontsize + 4)
    ax.set_title(title)
    if len(c_diff) > 0:
        ax.legend(loc='best')
    fig.tight_layout()
    return FigureRef(ax)


# --------------------------------------- Table formatting functions
def tab_correlation_matrix(cov, item_trait_map, name=None):
    """Normalized correlation matrix between IRT traits corresponding to items.
    :param cov: estimated, possibly un-normalized, covariance matrix,
        possibly including traits that have NO item correspondence
    :param item_trait_map: 2D boolean array mapping trait to item
        item_trait_map[i, t] == True <=> t-th trait determines i-th item response
    :param name: (optional) string file name of table
    :return: string with tabulated correlation values by trait = IOI-HA item
    """
    if name is None:
        name = 'Corr'
    trait_sel = np.any(item_trait_map, axis=0)
    c = cov[:, trait_sel][trait_sel, :]
    std = np.sqrt(np.diag(c))
    # Normalize covariance matrix
    c /= std
    c /= std[:, None]
    n_traits = c.shape[0]
    trait_labels = _make_trait_labels(item_trait_map)
    # *** use externally defined trait labels? ******
    align = 'r | ' + (n_traits-1) * 'r '
    header = ['Trait'] + trait_labels[1:]
    rows = []
    for i in range(n_traits-1):
        rows += [[trait_labels[i]] + i * [' '] + [f'{c[i, j]:.3f}'
                                                  for j in range(i+1, n_traits)]]
    return TableRef(_make_table(header, rows, align), name=name)


def tab_credible_diff(diff,
                      x_labels,
                      x_label='Population'):
    """create table with credible trait differences between populations
    represented by included group data sets
    :param diff: list of tuples ((i, j), p), indicating that
        x_labels[i] > x-labels[j] with joint credibility p, meaning
        prob{ x_labels[i] > s_labels[j] } AND all previous pairs } == p
    :param x_labels: list of category labels of compared data
    :param x_label: label of category for the difference
    :return: TableRef object with header lines + one line for each credible difference,
    """
    if len(diff) == 0:
        return None
    align = 'l l c l r'
    h = ['', x_label, '$>$', x_label, FMT['credibility']]
    rows = []
    col_0 = ''
    # ((i,j), p) = diff[0]  # separate format for first line
    for ((i, j), p) in diff:
        rows.append([col_0, x_labels[i], '$>$', x_labels[j], f'{100*p:.1f}\%'])
        col_0 = 'AND' # for all except first row
    return TableRef(_make_table(h, rows, align))


# ------------------------------------------ internal help functions:
def _nice_ticks(y_min, y_max):
    """Make nice sequence of equally spaced ticks
    with at most one decimal
    :param y_min: scalar axis min
    :param y_max: scalar axis max
    :return: t = array with tick values
        all(y_min <= t <= y_max
    """
    d = y_max - y_min
    if d < 1.:
        step = 0.1
    elif d < 2.:
        step = 0.2
    elif d < 5.:
        step = 0.5
    else:
        step = 1.
    return np.arange(np.ceil(y_min / step), np.floor(y_max / step) + 1) * step


def _x_tick_style(labels):
    """Select xtick properties to avoid tick-label clutter
    :param labels: list of tick label strings
    :return: dict with keyword arguments for set_xticklabels
    """
    maxL = max(len(l) for l in labels)
    rotate_x_label = maxL * len(labels) > 60
    if rotate_x_label:
         style = dict(rotation=25, horizontalalignment='right')
    else:
        style = dict(rotation='horizontal', horizontalalignment='center')
    # if len(labels) > 20:
    #     style['fontsize'] = 8
    # else:
    #     style['fontsize'] = 10
    return style


table_begin = {'latex': lambda align: '\\begin{tabular}{' + align + '}\n',
               'tab': lambda align: ''}
table_head_sep = {'latex':'\hline\n',
                  'tab':''}
table_cell_sep = {'latex': ' & ',
                  'tab':' \t '}
table_row_sep = {'latex': '\\\\ \n',
                 'tab': '\n'}
table_end = {'latex':'\hline\n\end{tabular}',
             'tab': ''}


def _make_table(header, rows, col_alignment):
    """Generate a string with table text.
    :param header: list with one string for each table column
    :param rows: list of rows, where
        each row is a list of string objects for each column in this row
    :param col_alignment: list of alignment symbols, l, r, or c
        len(col_alignment) == len(header) == len(row), for every row in rows
    :return: single string with complete table
    """
    def make_row(cells, fmt):
        return table_cell_sep[fmt].join(f'{c}' for c in cells) + table_row_sep[fmt]
    # ------------------------------------------------------------------------

    fmt = FMT['table_format']  # module global constant
    t = table_begin[fmt](col_alignment)
    t += table_head_sep[fmt]
    t += make_row(header, fmt)
    t += table_head_sep[fmt]
    t += ''.join((make_row(r, fmt) for r in rows))
    t += table_end[fmt]
    return t


def _make_trait_labels(item_trait_map):
    """Prepare trait labels for plots and tables
    :param item_trait_map: 2D boolean array mapping trait for each item
    :return: list of strings
    """
    # *** should be done externally ? ***
    trait_sel = np.any(item_trait_map, axis=0)
    item_trait_map = item_trait_map[:, trait_sel]
    n_items = len(item_trait_map)
    t_items = [np.arange(1, n_items + 1, dtype=int)[t_i]
               for t_i in item_trait_map.T]
    return ['Q(' + ','.join(f'{i}' for i in t_i) + ')'
            for t_i in t_items]


def _clean_file_name(s):
    """Make a string that can be used as file name
    :param s: string
    :return: clean_s,
        with whitespace replaced by _,
        and not-allowed chars removed
    """
    clean_s = s.replace(' ', '_')
    return clean_s.translate({ord(c): None for c in '(),.'})


# ------------------------------------------------------
if __name__ == '__main__':
    print(_clean_file_name('asd asdf. (asdf), asdf'))


"""This module defines help classes to access responses to questionnaire items.
Data may be stored in various file formats.

*** Class Overview:

ItemResponseDataSet: container for all response data for selected group(s),
    to be used as input for statistical analysis.
    Response data input must be an iterable over response records.
    A Response Record is a 1D list with integer-coded responses from ONE subject,
    with one ordinal integer value for each Questionnaire Item.
    The input data for one group may be
    1) a single list of subject response lists,
    2) a single table_reader.Table instance, reading from one data file,
    3) a single table_reader.Tables instance, chaining input from several iterable objects.

Questionnaire: description of a questionnaire, usually with several items

Item: description of a single item (question and answer alternatives)

*** Main File Interface Function:

item_response_table: general function interface to ONE data file in any allowed format

*** Input File Formats:

xlsx: Data can be imported from Excel workbook (xlsx) files,
    with data stored in ONE worksheet.
    Item response data are stored in specified columns,
    as defined by keyword parameter 'fields' to the Table reader class.
    See module table_reader_xlsx for details.

sql: Data can be imported from sql database files.
    Item response data are stored in specified database fields,
    as defined by keyword parameter 'fields' to the Table reader class.
    See module table_reader_sql for details.
    (table_reader_sqlite3 is no longer used.)

csv: Data can be imported from a csv text file.
    Item response data are stored in specified columns,
    identified by 'fields' string labels in the first line of the file.
    See module table_reader_csv for details.

spss: Data can be imported from an SSPS (.sav) file.
    See module table_reader_spss for details.
    However, SSPS reader might not work with newer python versions.
    May be easier to first convert the file to .csv format


*** Usage Example:

q_file = string or Path identifying questionnaire text (utf8) file
r_file = string or Path identifying response file

group0_file = item_response_table(r_file1, fmt='xlsx')
group1A_file = item_response_table(r_file1A, ...)  another data file
group1B_file = item_response_table(r_file1B, ...)  another data file
etc.

ids = ItemResponseDataSet.load(questionnaire=Questionnaire.load(q_file),
        groups={'Group0': group0_file,
                'Group1': Tables(group1A_file, group1B_file,...)
                }
        )

This data-set can then be used as input to create an analysis model, e.g.,
irm = OrdinalItemResponseModel.initialize(ids, n_traits=3)
irm.learn()
See run_irt.py for a complete example.

*** Version History:

2019-07-23, functional version, may need some cleanup
2019-07-29, new version with general table_reader interface
2019-08-25, minor cleanup
"""
# **** How to handle duplicate subjectID ??? ********

import numpy as np
from pathlib import Path

import logging
from importlib import import_module
from itertools import chain

ircPackage = 'ItemResponseCalc.'
# = top ref to sub-modules for selective imports

STORAGE_FMT = {'sqlite3': ircPackage + 'table_reader_sqlite',
               'sql': ircPackage + 'table_reader_sql',
               'xlsx': ircPackage + 'table_reader_xlsx',
               'spss': ircPackage + 'table_reader_spss',
               'sav': ircPackage + 'table_reader_spss',
               'csv': ircPackage + 'table_reader_csv'
               }
# = mapping of file format code to suitable table_reader module
# to be imported only if actually needed

logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------------
class Item:
    """Information describing ONE questionnaire item.
    """
    def __init__(self, question, responses, reverse=False):
        """
        :param question: string with explicit question text
        :param responses: list of strings, one for each possible response
        :param reverse: boolean = True if responses should be indexed in reverse order
            NOTE: Response data files may or may not already have recoded
            the responses by ordinal indices in the desired analysis order.
        """
        self.question = question
        self.responses = responses
        self.reverse = reverse

    def __repr__(self):
        return (self.__class__.__name__ + '(\n\t'
                + ',\n\t'.join((f'{k}={repr(v)}' for (k,v) in self.__dict__.items()))
                + ')')

    @property
    def n_response_levels(self):
        return len(self.responses)


class Questionnaire:
    """Container for a questionnaire with several items,
    required to specify number of items and number of response levels for each item,
    even if the actual questions and response alternatives are not included.
    """
    def __init__(self, header=None, items=None, item_response_levels=None):
        """
        :param header: (optional) string describing the questionnaire
        :param items: (optional) list of Item instances
        :param item_response_levels: (optional) necessary only if items is None:
            list with
            item_response_levels[i] = integer number of response levels for i-th item
        """
        self.header = header
        self.items = items
        if items is None:
            self.item_response_levels = item_response_levels
        else:
            self.item_response_levels = [i.n_response_levels for i in self.items]

    def __repr__(self):
        return (self.__class__.__name__ + '(\n\t'
                + ',\n\t'.join((f'{k}={repr(v)}' for (k,v) in self.__dict__.items()))
                + '\n\t)')

    @property
    def n_items(self):
        return len(self.item_response_levels)

    @classmethod
    def load(cls, file, sep='\t'):
        """Read questions and answers from given file
        :param file: string or Path object identifying input file
        :param sep: (optional) string, separator between response alternatives
        :return: cls instance
        """
        strip = ' \n'  # chars to strip away from end of input lines
        with Path(file).open('rt', encoding='utf8') as f:  # *********** other txt encodings ??? *****
            header = f.readline()
            items = []
            while True:
                q = f.readline()
                a = f.readline()
                reverse = f.readline()
                if a == '':
                    break  # even if q is non-empty
                else:
                    items.append(Item(question=q.rstrip(strip),
                                      responses=(a.rstrip(strip)).split(sep),
                                      reverse='rev' in reverse))
        return cls(header.rstrip(strip), items)


# ------------------------------------------------------------
class ItemResponseDataSet:
    """All result data for one complete item response analysis,
    including one or several groups of subjects.
    There should be at least one group with complete individual response data,
    otherwise the analysis cannot reveal correlations between items.
    """
    def __init__(self, questionnaire, groups):
        """
        :param questionnaire: single Questionnaire object
        :param groups: dict with elements (g_name, g_subjects), where
            g_name = string identifying the group
            g_subjects = iterable, e.g., a list or responses or an instance of a table_reader.Table subclass,
                yielding a sequence of subject response_records, one for each subject.
                response_record = a 1D list with
                response_record[i] = integer ordinal index of response to i-th item
                encoded with origin == 1.
                NOTE: different from python index origin!
                Missing responses MUST be indicated as 0.
                The Table object can apply user-supplied recoding functions,
                in case other response encodings are used in the file.
        """
        assert questionnaire is not None, 'Must have a Questionnaire object'
        self.questionnaire = questionnaire
        if groups is None:
            groups = dict()
        self.groups = groups

    def __repr__(self):
        return (self.__class__.__name__ + '('
                + f'\n\t questionnaire= {self.questionnaire.__class__.__name__} '
                + f'object with {self.questionnaire.n_items} items,'
                + '\n\t groups= {\n\t\t'
                + ',\n\t\t'.join(f'{g_key}: {g.__class__.__name__} with about {len(g)} rows'
                                 for (g_key, g) in self.groups.items()) + '}'
                + ')')

    @property
    def n_response_levels(self):
        return self.questionnaire.item_response_levels

    @property
    def n_items(self):
        return len(self.n_response_levels)

    @property
    def n_groups(self):
        return len(self.groups)

    @property
    def item_response_count(self):
        """Total number of response counts for each item,
        summed across all groups.
        :return: c = list of response count arrays, with
            c[i] = 1D array for i-th item,
            with responses encoded with origin == 0, i.e.,
            c[i][l] = count of response == l+1 for this item
        """
        c = [np.zeros(n_i, dtype=int) for n_i in self.n_response_levels]
        # counter, NOT including space for missing response coded as zero
        for (g_key, g_response) in self.groups.items():
            # g_response is iterable of item-response vectors, one for each subject,
            # with responses encoded with origin == 1 for the lowest ordinal response alternative
            for r_s in g_response:
                for (item, r_si) in enumerate(r_s):
                    if 1 <= r_si <= len(c[item]):  # i.e., real non-missing response
                        c[item][r_si - 1] += 1  # count first response alternative at index == 0
                    elif r_si > len(c[item]):
                        logger.warning(f'Group {repr(g_key)}: Response={r_si} for item {item} out of range')
                    # else just a missing data value, not counted
        return c


# ---------------------------------------------- File Interface factory:
def item_response_table(file_path, fmt=None, **kwargs):
    """Create interface to input file with individual item response data.
    :param file_path: Path or string identifying input file
    :param fmt: (optional) file format code, if not defined by file suffix
    :param kwargs: keyword arguments to table_reader.Table of selected format
    :return: a Table object representing the given file,
        which can be used as an iterator yielding subject response records.
    """
    file_path = Path(file_path)
    assert file_path.exists(), f'{file_path.name} not found'
    assert file_path.is_file(), f'{file_path.name} is not a file'
    if fmt is None:
        fmt = file_path.suffix[1:]
    if fmt in STORAGE_FMT:
        m = import_module(STORAGE_FMT[fmt])
        return m.Table(file_path, **kwargs)
    else:
        raise RuntimeError(f'Unknown file format: {repr(fmt)}')


# -------------------------------------------------------- Module TEST:
# if __name__ == '__main__':
#
#     def recode_missing_as_0(row):
#         for i in range(len(row)):
#             if row[i] is None:
#                 row[i] = 0
#     # --------------------------------------------
#
#     HAQ_PATH = Path.home() / 'Documents/LeijonKTH/heartech/HA-QualityRegistry/IRT-IOI-HA/IOI-HA-Data'
#     # ioiha_data_path = HAQ_PATH / 'IRT-IOI-HA' / 'IOI-HA-data'
#     # to ioi-ha data sets OTHER THAN the Swedish Quality Registry
#
#     print('*** Test Load Questionnaire:\n')
#     q = Questionnaire.load(HAQ_PATH / 'IOI-HA-English.txt')
#     print(q)
#     # ------------------------------------------------------------------
#
#     print('\n*** Testing table_reader_xlsx with Hickson data set:\n')
#
#     test_file = HAQ_PATH / 'Hickson' / 'Short_Eartrak AUS.xlsx'
#
#     irf = item_response_table(test_file,
#                               fields=[f'Q0{i}' for i in range(1, 8)],
#                               field_types=int,
#                               header_row=1,
#                               missing_values=['.', ' '])
#     print('printing a few records:')
#     for (i,r) in enumerate(irf):
#         print(r)
#         if i > 10:
#             break
#     print('printing first few records again:')
#     for (i,r) in enumerate(irf):
#         print(r)
#         if i > 5:
#             break
#
#     all_r = [r for r in irf]
#     print('number of records=', len(all_r))
#     print(f'len(irf) = ', len(irf))
#
#     # check subject duplicates:
#     irs = item_response_table(test_file,
#                               fields=['A'],
#                               header_row=0)
#     subjects = set()
#     dupl_subjects = set()
#     dupl_records = []
#     for (i,r) in enumerate(irs):
#         if r[0] in subjects:
#             print(f'Duplicate subject ID={repr(r[0])} found in row {i+2}')
#             dupl_subjects.add(r[0])
#             dupl_records.append(i)
#         else:
#             subjects.add(r[0])
#     print(f'Found {len(subjects)} unique subject IDs')
#     print(f'Found {len(dupl_subjects)} duplicate subject IDs in {len(dupl_records)} records')
#     print(dupl_subjects)
#
#     # ----------------------------------------- Use Hickson xlsx in data set:
#     print('\n*** Testing usin Hickson file in an ItemResponseDataSet object:\n')
#
#     irf = item_response_table(test_file,
#                               fields=[f'Q0{i}' for i in range(1, 8)],
#                               field_types=int,
#                               header_row=1,
#                               missing_values=['.', ' '],
#                               recode_fcns=recode_missing_as_0)
#
#     ids = ItemResponseDataSet(questionnaire=Questionnaire(item_response_levels=[5,5,5,5,5,5,5]),
#                               groups={'HicksonEarTrak': irf})
#
#     print(ids)
#     print('response_count=\n', ids.item_response_count)
#     print('total counts: ', [np.sum(c_i) for c_i in ids.item_response_count])
#
#     # ----------------------------------------- test Kramer SPSS data set:
#
#     print('\n*** Testing table_reader_spss with Kramer data set:\n')
#
#     kramer_file = item_response_table(file_path=HAQ_PATH / 'Kramer' / 'moeder_final - kopieforArne.sav',
#                                       fmt='spss',
#                                       fields=[f'ioi_v{i}' for i in range(1, 8)],
#                                       field_types=int,
#                                       missing_values=[9.0],
#                                       recode_fcns=recode_missing_as_0)
#
#     print(f'Using kramer_file with {len(kramer_file)} records')
#     print('printing a few records:')
#     for (i, r) in enumerate(kramer_file):
#         if i > 10:
#             break
#         print(r)
#
#     kramer_file2 = item_response_table(file_path=HAQ_PATH / 'Kramer' / 'NLSH-Kramer-T2IOI-HA.sav',
#                                        fmt='spss',
#                                        fields=[f'ioi_v{i}' for i in range(1, 8)],
#                                        field_types=int,
#                                        missing_values=[9.0],
#                                        recode_fcns=recode_missing_as_0)
#
#     print(f'Using kramer_file with {len(kramer_file2)} records')
#     print('printing a few records:')
#     for (i, r) in enumerate(kramer_file2):
#         if i > 10:
#             break
#         print(r)
#
#     ids = ItemResponseDataSet(questionnaire=Questionnaire(item_response_levels=[5,5,5,5,5,5,5]),
#                               groups={'Kramer': kramer_file})
#
#     print(ids)
#     print('response_count=\n', ids.item_response_count)
#     print('total counts: ', [np.sum(c_i) for c_i in ids.item_response_count])
#
#
# # ------------------------------------------------------- Check using Swedish data set:
#     print('\n*** Testing table_reader_sqlite with haq data set:\n')
#
#     def accept_record(r):
#         """Inclusion_crit for accepting a record
#         :param r: dict with all record fields
#         :param f: list of record field keys to be used
#         :return: boolean == True if record is acceptable
#         """
#         n_missing = sum(r_i is None for r_i in r)
#         within_range = all((r_i is None) or 1 <= r_i <= 5
#                            for r_i in r)
#         return n_missing <= 3 and within_range
#     # -------------------------------------------------------------------
#
#     HAQ_PATH = Path.home() / 'Documents/LeijonKTH/heartech/HA-QualityRegistry/ircData'
#     db_file = str(HAQ_PATH / 'development.sqlite3')
#
#     SELECT_FROM = ("prescriptions where " +
#                     "Respondent = 't' " +
#                     "AND article_company IS NOT NULL " +
#                     "AND date >= '2016-01-01' AND date <= '2016-01-30' "
#                    )
#
#     haq = item_response_table(file_path=db_file,
#                               select_from=SELECT_FROM,
#                               fields=[f'Q{i}' for i in np.arange(3,10)],
#                               accept_row=accept_record,
#                               recode_fcns=recode_missing_as_0)
#     # *** missing data is already coded as 0
#
#     print(f'len(haq) = ', len(haq))
#     print('*** First few records selected for analysis:')
#     n_rec = 10
#     for r in haq:
#         print(r)
#         n_rec -= 1
#         if n_rec <= 0:
#             break
#
#     print('\n*** Testing ItemResponseDataSet with sqlite dataset:\n')
#     ids = ItemResponseDataSet(questionnaire=Questionnaire(item_response_levels=[5,5,5,5,5,5,5]),
#                               groups={'HAQregistry': haq})
#
#     print(ids)
#
#     print('response_count=\n', ids.item_response_count)
#     print('total counts: ', [np.sum(c_i) for c_i in ids.item_response_count])
#
#     # -------------------------------------------------------------------
#     print('\n*** Testing ItemResponseDataSet with three sources:\n')
#
#     ids3 = ItemResponseDataSet(questionnaire=Questionnaire(item_response_levels=[5,5,5,5,5,5,5]),
#                                groups={'AU': irf,
#                                                   'NL': kramer_file,
#                                                   'SE': haq})
#
#     print(ids3)
#
#     print('response_count=\n', ids3.item_response_count)
#     print('total counts: ', [np.sum(c_i) for c_i in ids3.item_response_count])

# ------------------------------------------- plot response distributions



"""Interface to read tabular data stored in an sqlite3 database
e.g., the Swedish Quality Registry, provided by Peter Nordqvist, Hearing Bridge, Sweden.

*** MAIN CLASS:
Table --- iterable interface to an sqlite database,
    accessing fields by their names as used in the database.

NOTE: the database file is always accessed in READ-ONLY mode!

VERSION HISTORY:
2019-07-11, first functional version
2019-07-28, change to Table subclass
"""
# **** use sqlite3.Rows instead of row_dict ??? ***************

# ***** Can accept_row fcn detect and skip duplicate subjectID records ??? ***********

import numpy as np
# # from collections import Counter, OrderedDict
# import warnings

import sqlite3

from ItemResponseCalc import table_reader

# import logging
# logger = logging.getLogger(__name__)


# ---------------------------------------------------------------

class Table(table_reader.Table):
    """Table interface to sqlite3 database.
Provides generator function for all user records,
User can select which database records to include for analysis.
User can select which record fields are to be returned for analysis.
"""
    def __init__(self,
                 file_path,
                 select_from,
                 fields=None,
                 **kwargs):
        """
        :param file_path: Path or string with path to sqlite3 database file
        :param select_from: string to fill SELECT statement for records to include, e.g.,
            select_from = "prescriptions WHERE Respondent = 't' AND date >= '2016-01-01' "
            will find records using the SQL statement
            "SELECT * from prescriptions WHERE Respondent = 't' AND date >= '2016-01-01' "
        :param fields: list of key strings for selected fields of each database record
        :param kwargs: additional keyword properties for super-class, e.g.,
            field_types, keyed, accept_row, missing_values, recode_fcns
        """
        super().__init__(table_id=file_path, fields=fields, **kwargs)
        self.select_from = select_from

    def __iter__(self):
        """Generator for all records SELECTed in property select_from,
        AND also accepted and pre-processed
        as defined by self.accept_row, self.recode_fcns,
        yielding list with selected field values from those records.
        """
        con = sqlite3.connect('file:' + str(self.table_id) + '?mode=ro', uri=True)
        # *** NOTE: READ ONLY!
        con.row_factory = sqlite3.Row
        query = 'SELECT * FROM ' + self.select_from
        for (rec_n, rec) in enumerate(con.execute(query)):
            if rec_n % self.sample_factor == 0:
                if self.fields is None:
                    r = list(rec)
                else:
                    r = [rec[f] for f in self.fields]
                r = self.process_row(r)
                if r is not None:
                    yield r
        con.close()
        # ******** must close manually, even though connector is environment manager!

    def __len__(self):
        """Total number of records, disregarding accept_row
        :return: scalar integer
        """
        con = sqlite3.connect('file:' + str(self.table_id) + '?mode=ro', uri=True)
        # con.row_factory = sqlite3.Row
        # count = 'COUNT(*)'
        query = "SELECT COUNT(*) FROM " + self.select_from
        count_cursor = con.execute(query)
        n_records = count_cursor.fetchone()[0]
        con.close()
        return n_records // self.sample_factor

    def preview(self, max_records=10):
        """Generator for records SELECTed in property select_from,
        without any further filtering or recoding.
        """
        con = sqlite3.connect('file:' + str(self.table_id) + '?mode=ro', uri=True)
        # *** NOTE: READ ONLY!
        con.row_factory = sqlite3.Row
        query = 'SELECT * FROM ' + self.select_from
        # result = []
        for (rec_n, rec) in enumerate(con.execute(query)):
            if rec_n > max_records:
                break
            # result.append(dict(rec))
            yield dict(rec)
        con.close()
        # ******** must close manually, even though connector is environment manager!
        # return result

# ----------------------------------- Internal help functions:

# def row_as_dict(cursor, row):  # sqlite3.Row can also return fields by name
#     """row_factory for sqlite connector object, reading each row as a dict
#     """
#     d = {}
#     for idx, col in enumerate(cursor.description):
#         d[col[0]] = row[idx]
#     return d


# ---------------------------------------------- Module TEST:



"""Script to display analysis results for ordinal data,
using an OrdinalItemResponseModel object trained and saved earlier,
by script like run_irt.py

This script produces tables and figures
that can easily be imported into a manuscript for publication.

2019-07-21, OK version including standardization result
2019-08-25, adapted for model with automatic selection of traits
2020-06-06, updated for standalone ItemResponseCalc package
"""
import numpy as np
import pickle
from pathlib import Path

import logging
from ItemResponseCalc import ir_logging

from ItemResponseCalc import ir_display as ird


# ------------------------------------------------------------------------

TOP_PATH = Path.home() / 'Documents/LeijonKTH/heartech/HA-QualityRegistry'
TOP_PATH = TOP_PATH / 'IRT-IOI-HA'
# change this to top-dir for everything to be read or written here

result_dir = TOP_PATH / 'IOI-HA-results'
# change this to same as in run_irt
# ------------------------------------------------------------------------

run_dir = '2020-05-28-1612'
# change this to the saved result from run_irt

save_dir = result_dir / run_dir

model_file = 'irt-model.pkl'
# = name of saved model generated by run_irt

log_file = 'run_show_log.txt'
# = name of log file saved in same sub-folder
# -------------------------------------------------------------------------

# ------------------------------------- setup logging:
ir_logging.setup(save_dir / log_file)
# -----------------------------------------------------------

logging.info('Showing IRT model results in ' + str(save_dir))

with (save_dir / model_file).open('rb') as f:
    irm = pickle.load(f)

logging.info('Learned Model =\n' + repr(irm))
itm = irm.item_trait_map()

# ------------------------------------- Trait Covariance before standardization
# only informative, may be commented out

c = irm.precision_within.mean_inv
logging.info('Predictive Trait Covariance Within Populations =\n'
             + np.array2string(c, precision=3, suppress_small=True))

c = irm.precision_among.mean_inv
logging.info('Predictive Trait Variance Among Populations= '
             + np.array2string(c, precision=3, suppress_small=True))

# ------------------------------------- Global Predictive Covariance
c = irm.predictive_individual_cov()
logging.info('Predictive Global Trait Cov (before standardization) =\n'
             + np.array2string(c, precision=3, suppress_small=True))

# -------------------------------------- prune and standardize:
irm.prune()  # keep only traits that were really needed to model the data

irm.standardize()  # re-scale all parameters for unity global trait variance
c = irm.predictive_individual_cov()
logging.info('Pruned and Standardized Predictive Global Trait Cov =\n'
             + np.array2string(c, precision=3, suppress_small=True))
s = np.sqrt(np.diag(c))
logging.info('Predictive St.Dev. after Standardization = '
             + np.array2string(s, precision=3))

# ------------------------------------- display with standardized scale
ir_display = ird.ItemResponseDisplaySet.show(irm,  # ***** define trait labels here ? ****
                                             mapping_item=-1,
                                             figure_format='pdf',  # OR eps, png,...
                                             table_format='latex',  # OR tab
                                             )
ir_display.save(save_dir)

# ------------------------------------- Within-group trait correlation
if irm.n_traits > 1:
    cov_within = irm.precision_within.mean_inv
    logging.info('Predictive Trait Covariance Within Populations:\n' +
                 np.array2string(cov_within, precision=3))
    s = np.sqrt(np.diag(cov_within))
    c = cov_within / s
    c /= s[:, np.newaxis]
    # = normalized correlation matrix
    logging.info('Predictive Trait Correlation Within Populations:\n' +
                 np.array2string(c, precision=3))

    # ------------------------------------- PCA of within-group covariance
    (eig_val, eig_vec) = np.linalg.eigh(cov_within)
    logging.debug('Cov eig_val= ' +
                 np.array2string(eig_val, precision=3))
    logging.debug('Cov eig_vec=\n' +
                 np.array2string(eig_vec, precision=3))
    eig_val_cum = np.cumsum(eig_val[::-1]) / np.sum(eig_val)
    logging.info(f'One, Two, etc Principal Trait Factors: ' +
                 ', '.join(f'{ev:.1%}'
                           for ev in eig_val_cum) +
                 ', of Variance within Populations')
    # ***** Check factor analysis with varimax rotation ???

logging.shutdown()



"""This module defines classes representing the predictive trait distributions
for an item_response_model.OrdinalItemResponseModel instance,
trained on subject responses.
"""
import numpy as np

RANDOM = np.random.default_rng()
# = module-global random generator, Numpy 1.17


class GaussianGivenParam:
    """Gaussian distribution of a vector-valued random variable
    with parameter vector mu and precision matrix tau,
    which may be either fixed numpy arrays or random variables.

    Used for PREDICTIVE multivariate distributions
    that may be an infinite mixture of Student-t distributions,
    which has no known exact form.

    Nevertheless,
    this class can draw random variates from the exact predictive distribution,
    and calculate a good approximation to its covariance.
    """
    def __init__(self, mu, tau):
        """
        :param mu: 1D array or random-variable instance for the mean
        :param tau: 2D array or random-variable instance for the precision matrix
        """
        self.mu = mu
        self.tau = tau

    def __repr__(self):
        return (self.__class__.__name__ + '('
                + '\n\t mu= ' + self.mu.__class__.__name__ + ','
                + '\n\t tau= ' + self.tau.__class__.__name__
                + ')')

    def cov(self):
        """Expected covariance matrix,
        averaged across parameter distributions
        """
        if isinstance(self.tau, np.ndarray):
            c = np.linalg.inv(self.tau)
        else:
            c = np.linalg.inv(self.tau.mean_inv)
        # c = expected conditional cov, given mu
        if isinstance(self.mu, np.ndarray):
            return c
        else:
            return self.mu.cov + c

    def rvs(self, size=1):
        """Random vectors drawn from self,
        based on random samples of parameters
        :param size: integer number of desired samples
        :return: 2D array x, with
            x[n, :] = n-th sample drawn from self

        Method: scipy.stats.multivariate_normal can only use 1D mean and 2D cov,
        so we must do it directly using eigen-analysis of the precision matrix
        """
        if isinstance(self.mu, np.ndarray):
            m = self.mu
        else:
            m = self.mu.rvs(size=size)
        if isinstance(self.tau, np.ndarray):
            (e_val, e_vec) = np.linalg.eigh(self.tau)
        else:
            (e_val, e_vec) = np.linalg.eigh(self.tau.rvs(size=size))
        # random deviations with zero mean:
        if np.isscalar(size):
            z_shape = (size, m.shape[-1])
        else:
            z_shape = (*size, m.shape[-1])
        z = RANDOM.standard_normal(size=z_shape)
        # = standard normal independent random numbers
        return m + np.einsum('...i, ...j, ...ji -> ...j',
                             z, 1. / np.sqrt(e_val), e_vec)


# -------------------------------------------------------- TEST:
if __name__ == '__main__':
    from scipy.stats import multivariate_normal as mnorm

    # Testing some module functions

    m = np.array([1.,2.,3., 4.])
    x = mnorm.rvs(mean=m, cov=np.eye(3))
    print('x.shape= ', x.shape)
    print('x=\n', x)


import setuptools
import ItemResponseCalc

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ItemResponseCalc",
    version= ItemResponseCalc.__version__,
    author="Arne Leijon",
    author_email="leijon@kth.se",
    description="Statistical Analysis of Questionnaire Response Data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    license='MIT',
    classifiers=(
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Multimedia :: Sound/Audio"
    ),
    keywords = 'questionnaire item-response-theory Bayesian',
    install_requires=['numpy>=1.17', 'scipy', 'matplotlib', 'samppy', 'openpyxl'],
    python_requires='>=3.6'
)



"""Interface to read tabular data from a file,
e.g., a flat database or an Excel sheet or csv file.
This file is a base module defining a superclass for all types of data files.

*** Main Classes
Table --- superclass for variants of concrete readers of tabulated data.

Tables --- iterable chaining several sub-iterables with tabulated data.
"""
# **** use reader as context manager?

from itertools import chain

import logging

logger = logging.getLogger(__name__)


# ---------------------------------------------------------------
def default_accept_all(r):
    """Default for Table property accept_row.
    :param r: list with one record from database
    :return: always True, without any checking
    """
    return True


# ---------------------------------------------------------------
class Table:
    """Superclass for all types of concrete table-storage access methods,
    allowing access to files with data stored in rows and columns.
    Typical usage:
        table = Table(...args)
        for row in table:
            do something with the row
   """
    def __init__(self, table_id,
                 fields=None,
                 field_types=int,
                 # keyed=False,
                 missing_values=None,
                 sample_factor=1,
                 accept_row=None,
                 recode_fcns=None):
        """
        :param table_id: string, file path, or other object defining concrete storage for reading
            as interpreted by sub-class
        :param fields: sequence of labels for fields to be returned in each record
            if None, ALL fields are returned
        :param field_types: list of required value types, or a single such type
            if single type, it is applied for ALL returned field values
            NOTE: MUST be callable objects that allow direct type cast, e.g.,
            field_types = int; will be used as int(row[field])
        :param keyed: boolean, if True: yield record as dictionary, otherwise as a list
        :param missing_values: list of field values as stored in the file, to be treated as missing data
            replaced by None before output.
            Field values are tested before any recoding or type-cast.
        :param sample_factor: (optional) scalar integer for down-sampling among records
            Only one record is returned for every sample_factor records in the file
        :param recode_fcns: (optional) function or list of functions to be applied to each record before output
        :param accept_row: (optional) single function to determine if a record is to be accepted
        NOTE: test for missing values, type-cast, and recoding are applied in the order
            defined in method process_row.
        """
        self.table_id = table_id
        self.fields = fields
        # ******** allow duplicates in self.fields ???
        if field_types is not None and type(field_types) is not list:
            field_types = len(self.fields) * [field_types]
        self.field_types = field_types
        if missing_values is None:
            missing_values = []
        self.missing_values = missing_values
        if accept_row is None:
            accept_row = default_accept_all
        self.accept_row = accept_row
        if recode_fcns is None:
            recode_fcns = []
        if type(recode_fcns) is not list:
            recode_fcns = [recode_fcns]
        self.recode_fcns = recode_fcns
        self.sample_factor = sample_factor

    def __repr__(self):
        return (self.__class__.__name__ + '(\n\t'
                + ',\n\t'.join((f'{k}={repr(v)}' for (k,v) in self.__dict__.items()))
                + ')')

    def __len__(self):
        """Number of (down-sampled) data records in file,
        not including header,
        exact or crudely estimated.
        This default method simply reads and counts all rows.
        Subclasses may implement faster method.
        :return: scalar integer
        """
        n = 0
        for _ in self:
            n += 1
        return n

    def __iter__(self):
        """Must be implemented by subclass
        :return: iterator over rows
        """
        raise NotImplementedError

    def process_row(self, row):
        """Check and convert ONE row as required by self properties
        :param row: list with elements corresponding to self.fields
        :return: row modified in place, or None if not accepted
        """
        # ****** raise NotAcceptedError if any error, using try-except in __iter__ ???
        self._check_missing(row)
        try:
            self._type_cast(row)
        except ValueError:
            logger.warning(f'Type error in {row}; not included')
            return None
        for fcn in self.recode_fcns:
            if fcn is not None:
                fcn(row)
        if self.accept_row(row):
            return row
        else:
            return None

    # def preview(self, max_records=10):  # ************* needed ?
    #     """Generator of raw data rows from the file.
    #     Must be implemented by sub-class.
    #     :param max_records: limit on the number of records
    #     :param keyed: boolean: if True, return data as dict  # ************ ???
    #     :return: iterator over data rows, as stored without any type casting or other checks
    #     """
    #     raise NotImplementedError

    # ----------------------------------------------------- Private:
    def _check_missing(self, row):
        """Check and convert any missing data in row
        :param row: list with selected field values from one record
        :return: None, row elements recoded in place
        """
        for i in range(len(row)):
            if row[i] in self.missing_values:
                row[i] = None

    def _type_cast(self, row):
        """Change type of row elements if required,
        change to None, if desired type-cast was not possible.
        :param row: list with selected field values from one record
        :return: None, row elements recoded in place
        """
        if self.field_types is not None:
            for (i, (r_i, type_i)) in enumerate(zip(row, self.field_types)):
                if r_i is not None:
                    try:
                        row[i] = type_i(r_i)
                    except ValueError:
                        logger.warning(f'Cannot convert {repr(r_i)} to {type_i} in {row}')
                        row[i] = None
    # ---------------------------------------------------------------


class Tables:
    """Iterable chaining sub-iterables
    """
    def __init__(self, *table_list):
        """
        :param table_list: list of iterables, e.g., lists or table_reader.Table objects.
        """
        self.table_list = table_list

    def __len__(self):
        """Total number of data records in all included sub-tables,
        exact or crudely estimated.
        :return: scalar integer
        """
        return sum(len(t) for t in self.table_list)

    def __iter__(self):
        """iterator of records in all included sub-tables
        :return: iterator over response records
        """
        return chain(*self.table_list)




"""Interface to read tabular data stored in MySQL database

*** MAIN CLASS:
Table --- iterable interface to an sqlite database,
    accessing fields by their names as used in the database.


NOTE: the database file is always accessed in READ-ONLY mode!

VERSION HISTORY:
2020-02-02, first functional version
"""
import numpy as np

import mysql.connector

from ItemResponseCalc import table_reader

# import logging
# logger = logging.getLogger(__name__)


# ---------------------------------------------------------------

class Table(table_reader.Table):
    """Table interface to sqlite3 database.
Provides generator function for all user records,
User can select which database records to include for analysis.
User can select which record fields are to be returned for analysis.
"""
    def __init__(self,
                 table_id,
                 connect_kwargs,
                 fields=None,
                 select_where='',
                 select_args=(),
                 **kwargs):
        """
        :param table_id: name of table in the SQL database
        :param connect_kwargs: dict with all input arguments for mysql.connect()
        :param select_where: string part of SELECT statement for records to include, e.g.,
            select_where = "Respondent = 1 AND date >= '2016-01-01' "
        :param select_args: (optional) tuple with arguments to select_where format string
        :param fields: list of key strings for selected fields of each database record
        :param kwargs: additional keyword properties for super-class, e.g.,
            field_types, keyed, accept_row, missing_values, recode_fcns
        """
        super().__init__(table_id=table_id, fields=fields, **kwargs)
        self.connect_kwargs = connect_kwargs
        self.select_where = select_where
        self.select_args = select_args

    def __iter__(self):
        """Generator for all records SELECTed in select_where, select_args
        AND also accepted and pre-processed
        as defined by self.accept_row, self.recode_fcns,
        yielding each row as a list with values of fields specified in self.fields
        """
        if self.fields is None:
            self.fields = '*'  # = all in database
        query = ('SELECT ' + ', '.join(self.fields)
                 + ' FROM ' + self.table_id
                 + ' WHERE ' + self.select_where)
        # *********** set time-out if connect fails ********************
        cnx = mysql.connector.connect(**self.connect_kwargs)
        cur = cnx.cursor()
        cur.execute(query, self.select_args)
        for (rec_n, rec) in enumerate(cur):
            if rec_n % self.sample_factor == 0:
                r = self.process_row(list(rec))  # ***************
                # r = rec
                if r is not None:
                    yield r
        cur.close()
        cnx.close()

    def __len__(self):
        """Total number of records, disregarding accept_row
        :return: scalar integer
        """
        query = ('SELECT COUNT(*) FROM '  + self.table_id
                 + ' WHERE ' + self.select_where)
        # *********** set time-out if connect fails ********************
        cnx = mysql.connector.connect(**self.connect_kwargs)
        cur = cnx.cursor()
        cur.execute(query, self.select_args)
        n_records = cur.fetchone()[0]
        cur.close()
        cnx.close()
        return n_records // self.sample_factor

    # def preview(self, max_records=10):  # ******** needed ?
    #     """Generator for records SELECTed in property select_where,
    #     without any further filtering or recoding.
    #     """
    #     query = ('SELECT * FROM ' + self.table_id
    #              + ' WHERE ' + self.select_where)
    #     cnx = mysql.connector.connect(**self.connect_kwargs)
    #     cur = cnx.cursor()
    #     cur.execute(query, self.select_args)
    #     column_names = cur.column_names
    #     for (rec_n, rec) in enumerate(cur):
    #         if rec_n > max_records:
    #             break
    #         # result.append(dict(rec))
    #         yield dict((k, v) for (k, v) in zip(column_names, rec))
    #     cur.fetchall()  # remaining
    #     cur.close()
    #     cnx.close()
    #     # return result


# ---------------------------------------------- Module TEST:

# if __name__ == '__main__':
#     import datetime as dt
#
#     haq_config = dict(user='leijon',
#                       password='*********',
#                       host='172.16.0.7',
#                       port='1111',
#                       database='hbdb')
#     table = 'prescriptions'
#
#     SELECT_WHERE = ("respondent = 1 " +
#                     "AND article_company IS NOT NULL " +
#                     "AND date BETWEEN %s AND %s"
#                     )
#     where_args = (dt.date(2019, 1, 1), dt.date(2019, 12, 31))
#
#     fields = ['id', 'civil_registration_number'] + [f'Q{i}' for i in np.arange(3, 10)]
#     ioiha_query = 'SELECT ' + ', '.join(fields) + ' FROM ' + table + ' where ' + SELECT_WHERE
#     print('query: ', ioiha_query)  # , where_args)
#
#     # --------------------------------- Test Table
#     haq_table = Table(table_id=table,
#                       connect_kwargs=haq_config,
#                       select_where=SELECT_WHERE,
#                       select_args=where_args,
#                       fields=fields)
#
#     # print('len(haq_table)= ', len(haq_table))
#     print('\n*** haq_table.preview(10)=\n')
#     for r in haq_table.preview(10):
#         print(r)
#



""" Module defining probability distribution of rating-interval thresholds for
response categories in an OrdinalItemResponseModel,
implementing the IRT Graded Response Model.

*** Main Class:
OrdinalItemScale --- defining response intervals for one item
    represented by IRT Graded Response Model.
    Response intervals are defined by a monotonically increasing
    sequence of response thresholds.

Method:
In this version, response thresholds are indirectly determined by vector
eta = logarithm of non-normalized width parameters in the logistic-mapped domain.
The prior distribution of eta is Gaussian,
and the posterior is approximated by a set of equally probable sample vectors,
or by a single MAP estimate.

To avoid numerical overflow during sampling,
the eta samples are also restricted between (wide) fixed bounds,
so the prior eta distribution is actually a (mildly) truncated Gaussian.

2019-07-08 first version
2019-07-14, used HamiltonianBoundedSampler for numerical safety
2019-07-23, fixed bug in scale.initialize, Initial Trait Scale set externally
2019-08-10, use a TraitSelector object for item-trait mapping
2019-08-16, try uniform TraitSelector initialization
2019-08-24, use np.random.Generator for sampling
2019-08-25, sample_trait and typical_trait assuming Gaussian inter-individual trait distribution
2020-02-06, adapt including initial MAP point estimate for faster(?) sampling
2020-02-13, allow n_samples == 1 in method relative_entropy_re_prior
"""
# **** after testing, include latent_scale as OrdinalItemScale property
# **** allow string item_id ? *********************

import numpy as np
from scipy.stats import norm
from scipy.optimize import minimize

from .safe_logistic import logistic

from samppy import hamiltonian_sampler as ham
from samppy.sample_entropy import entropy_nn_approx as entropy

from .item_trait_selector import TraitSelector

import logging
logger = logging.getLogger(__name__)

# logger.setLevel(logging.INFO)
# logger.setLevel(logging.DEBUG)

ham.VECTOR_AXIS = 1  # just in case default has changed...

# ---------------------------- Module Default Constants:

PRIOR_ETA_SCALE = 3.  # 1.  # 3. ??? *******
# scale of Gaussian prior distribution of eta parameters
# defining a sufficiently diffuse prior

PSEUDO_COUNT = 0.5
# = Jeffreys prior for Dirichlet distribution of probability vector,
# given response counts

RANDOM_G = np.random.default_rng()
# = module-global random generator, Numpy 1.17


# ------------------------------------------- main initialization function:
def make_scales(item_response_count, trait_prob, n_scale_samples, trait_scale):
    """Initialize all item scale objects, and corresponding trait mapping objects.
    :param item_response_count: 1D list of response-count lists
        item_response_count[i][l] = number of responses == l for i-th item
    :param trait_prob: prior TraitProb instance initialized by caller
    :param n_scale_samples: scalar integer number of samples in each scale object
    :param trait_scale: scalar initially assumed scale of model trait values
    :return: scales = list with one OrdinalItemScale object for each item
    """
    scales = [OrdinalItemScale.initialize(count_i,
                                          n_samples=n_scale_samples,
                                          trait_scale=trait_scale,
                                          trait_prob=trait_prob,
                                          item_index=i)
              for (i, count_i) in enumerate(item_response_count)]
    return scales


# --------------------------------------------------------------------
class OrdinalItemScale:
    """Class defining response thresholds,
    and learning of scale parameters,
    for ONE item in an OrdinalItemResponseModel.

    Response interval widths in the logistic-transformed domain
    are proportional to 1D array w, which is represented as
    w = np.exp(eta)
    In each adapt step, the posterior distribution of eta
    is approximated by a set of sample vectors,
    or by a single MAP estimate.
    """
    @classmethod
    def initialize(cls, response_count, trait_scale, n_samples, trait_prob, item_index):
        """Initialize a scale object crudely to agree with given response counts.
        :param response_count: 1D array with total response counts
            response_count[l] = count of response == l
        :param trait_scale: scalar = initial scale of theta distribution
        :param n_samples: number of samples
        :param trait_prob: single prior TraitProb object for all scale.trait_sel objects
        :param item_index: scalar integer item identifier for this scale
        :return: a single cls object
        """
        f = np.cumsum(response_count + PSEUDO_COUNT)
        p_cum = f / f[-1]
        tau = norm.ppf(p_cum[:-1], scale=trait_scale)
        # tau[l] = upper limit of interval for response==l
        cum_w = np.concatenate(([0.], logistic.cdf(tau), [1.]))  # corrected 2019-07-23
        w = np.diff(cum_w)
        eta = np.log(w)
        # eta[l] = log width in transformed domain for l-th response interval
        eta = np.tile(eta, (n_samples, 1))
        return cls(eta, TraitSelector(trait_prob.mean), item_index)

    def __init__(self, eta, trait_sel, item_index):
        """
        :param eta: log interval width in logistic-transformed domain
            eta[n, l] = n-th sample of log-width for l-th response interval
        :param trait_sel: a TraitSelector object for this item
        :param item_index: scalar integer item identifier for this scale
        """
        self.item_index = item_index
        eta -= np.mean(eta, axis=1, keepdims=True)
        # forced to zero mean
        too_small = np.log(0.001)
        too_big = np.log(1000.)
        # => ratio about 1e6 between widest and smallest interval width
        b = [(too_small, too_big) for _ in range(eta.shape[-1])]
        # must restrict sample eta values to avoid numerical overflow
        self.sampler = ham.HamiltonianBoundedSampler(fun=neg_logprob,
                                                     jac=d_neg_logprob,
                                                     x=eta,
                                                     bounds=b,
                                                     epsilon=0.2,
                                                     n_leapfrog_steps=10)
        # sampler.args to be defined later
        # NOTE: eta is stored only as sampler.x
        self.trait_sel = trait_sel

    def __repr__(self):
        return (self.__class__.__name__ + f'(item_index= {self.item_index},'
                + f'\n\teta=\n\t{repr(self.eta)},'
                + f'\n\ttrait_sel= {repr(self.trait_sel)}'
                + ')')

    @property
    def eta(self):
        return self.sampler.x

    @property
    def n_samples(self):
        return self.sampler.x.shape[0]

    @property
    def n_response_levels(self):
        return self.sampler.x.shape[1]

    @property
    def tau(self):
        """Response-thresholds samples representing current distribution.
        :return: 2D array tau, with
            tau[m, l] = m-th sample of UPPER threshold for l-th interval,
            NOT including extreme values at -inf and +inf
            tau.shape[-1] == self.n_response_levels - 1
        """
        return _tau(np.exp(self.eta))

    def typical_trait(self, theta_scale=1.):
        """Point estimates of trait values, given any response,
        estimated at conditional mean within each response interval,
        assuming Gaussian traits with zero mean and scale = theta_scale,
        using current sampled distribution of scale thresholds.
        :param theta_scale: scalar = current (assumed) scale of theta distribution
        :return: theta = 2D array, with
            theta[m, l] = m-th sample of typical trait for response == l
            theta.shape[-1] == self.n_response_levels
        """
        # typ_median = _cond_median(self.tau, scale=theta_scale)
        typ_mean = _cond_mean(self.tau, scale=theta_scale)
        return typ_mean

    def sample_trait(self, r, n_samples, trait_scale=1.):
        """
        Generate samples of possible trait values, given subjects' responses
        :param r: 1D array with integer response indices, origin == 0
            r[s] = l in range(self.n_response_levels
            means s-th subject gave l-th response to this item
        :param n_samples: number of desired samples for each subject
        :param trait_scale: scalar = current (assumed) scale of theta distribution
        :return: th = 2D array of samples
            th[m, s] = m-th sample for s-th subject
        """
        def s_sample(r_s):
            """Sample in transformed (0, 1) range for ONE subject
            :param r_s: scalar integer response by this subject
            :return: 1D array th_s; len(th_s) == n_samples
            """
            if r_s < 0:
                return RANDOM_G.uniform(low=th_limits[0], high=th_limits[-1], size=n_samples)
            else:
                return RANDOM_G.uniform(low=th_limits[r_s], high=th_limits[r_s + 1], size=n_samples)
        # ----------------------------------------------------------------------
        th_limits = np.mean(self.tau, axis=0)
        th_limits = np.concatenate(([np.finfo(float).tiny],
                                    norm.cdf(th_limits, scale=trait_scale), [1.]))
        # tiny lowest limit, because RANDOM.uniform samples in [0, 1),
        # but we allow only samples in (0., 1)
        th = [s_sample(r_s) for r_s in r]
        return norm.ppf(np.array(th).T, scale=trait_scale)

    def ordinal_prob(self, theta):
        """Conditional probability of ordinal responses, given trait value,
        averaged across scale samples
        :param theta: array of trait values
        :return: pr = array of response probabilities
            pr[..., l] = P{response = l | theta[...]}
            = logistic.cdf(theta[...] - tau[l]) - logistic.cdf(theta[...] - tau[l-1])
            pr.shape == (*theta.shape, self.n_response_levels)
        """
        try:
            s = self.latent_scale
        except NameError:
            s = 1.
        t = self.tau - theta[..., np.newaxis, np.newaxis]
        t /= s
        cdf = np.mean(logistic.cdf(t), axis=-2)  # average across samples
        p = np.concatenate((np.zeros((*cdf.shape[:-1], 1)),
                            cdf,
                            np.ones((*cdf.shape[:-1], 1))),
                            axis=-1)
        p = np.diff(p, axis=-1)
        return p

    def mean_ordinal(self, theta):
        """Expected value of ordinal response level, given trait
        :param theta: array of trait values
        :return: r = expected "response" on continuous scale
            r[...] = sum_l l * P{response = l | theta[...]}

        NOTE: Since the ordinal responses are NOT on a linear interval scale,
        it is formally MEANINGLESS to take the mean of ordinal responses.
        Therefore, this result should be used only for illustration purpose,
        with caution about the scale non-linearity.

        Nevertheless, the expectation of ordinal responses
        is sometimes used without hesitation in the Rasch research tradition,
        Ref, e.g.: Wright and Masters (1982), Eq. 5.4.1.
        The expected rating plotted vs individual trait value
        is sometimes called Item Characteristic Curve (ICC).
        """
        return np.dot(self.ordinal_prob(theta), np.arange(self.n_response_levels))

    def rating_from_trait(self, theta):  # *** use mean_ordinal instead **********
        """Calculate mean equivalent continuous-valued "rating",
        in finite interval (- 0.5, n_rating_levels - 0.5)
        corresponding to given latent variable(s), such that
        th = tau[l] -> rating = l + 0.5
        with linear interpolation in the logistic-transformed (0, 1) range.
        :param theta: array-like 1D sequence of trait values in range (-inf, +inf)
        :return: z = rating-transformed values of y
            z.shape == y.shape
        """
        p_theta = logistic.cdf(theta)
        p_tau = logistic.cdf(np.mean(self.tau, axis=0))
        p_tau = np.concatenate(([0.], p_tau, [1.]))
        # = p corresponding to ratings [-0.5, 0.5, ..., n_rating_levels-0.5
        n_tau = len(p_tau)
        rp = np.linspace(-0.5, n_tau - 1.5, n_tau)
        return np.interp(p_theta, p_tau, rp) # + self.min_rating

    def adapt(self, groups, trait_prob):
        """One variational-inference update step of sample parameters eta.
        :param groups: list with ResponseGroup instances
            that can calculate log-likelihood values for given response thresholds.
        :param trait_prob: single TraitProb object, prior for self.trait_sel
        :return: self
            Must call relative_entropy_re_prior later !
        """
        logger.debug(self.__class__.__name__ + f'[{self.item_index}].adapt(...)')
        # ------ adapt trait selector first, with old tau:
        log_trait_resp = sum(np.mean(g.item_logprob_by_tau(self.tau, self.item_index),
                                     axis=0)
                             for g in groups)
        # log_trait_resp[t] = mean logprob given t-th trait
        self.trait_sel.adapt(log_trait_resp, trait_prob)

        # ------ adapt self.eta, to define new self.tau:
        self.sampler.args = (groups, self.item_index, self.trait_sel.mean)
        # ---------- first adjust by point estimate:
        eta_old = np.mean(self.sampler.x, axis=0)
        eta_new = self.adapt_point(eta_old)
        self.sampler.x += eta_new - eta_old
        logger.debug(f'scale[{self.item_index}] point adapted: diff= '
                     + np.array2string(eta_new - eta_old,
                                       precision=2,
                                       suppress_small=True))
        if self.n_samples == 1:
            logger.debug(self.__class__.__name__ + f'[{self.item_index}].adapt(...) MAP finished')
            return self
        # otherwise continue with sampling
        # --------------------------------
        max_steps = 20
        try:
            self.sampler.safe_sample(min_steps=5, max_steps=max_steps)
            if self.sampler.n_steps >= max_steps:
                logger.warning(f'Scale[{self.item_index}] Done {self.sampler.n_steps} = MAX allowed sampling steps')
            else:
                logger.debug(f'Scale[{self.item_index}] Done {self.sampler.n_steps} sampling steps')
            if self.sampler.accept_rate > 0.95:  # ****** do this in samppy.safe_sample ********
                self.sampler.epsilon *= 1.3
                logger.debug(f'Scale[{self.item_index}]: High sampler accept_rate; '
                             + 'increased epsilon = {self.sampler.epsilon}')
        except ham.AcceptanceError:
            logger.warning((f'Scale[{self.item_index}]: AcceptanceError: accept_rate= {self.sampler.accept_rate:.1%} '
                            + f'of {self.sampler.n_trajectories}; '
                            + f'epsilon reduced to {self.sampler.epsilon:.5f}'))
            # ****** keep going anyway ********
        else:
            logger.debug(f'Scale[{self.item_index}]: Sampler accept_rate = {self.sampler.accept_rate:.1%}')
            logger.debug(f'Scale[{self.item_index}]: Sampler epsilon = {self.sampler.epsilon:.5f}')
        self.sampler.x -= np.mean(self.sampler.x, axis=-1, keepdims=True)
        # adding a constant to all eta does not change any thresholds
        # --------------------------------- DONE distribution of eta
        if self.n_samples > 10:
            logger.debug(f'Scale[{self.item_index}]: tau percentiles=\n'
                         + np.array2string(np.percentile(self.tau, axis=0, q=[5, 50, 95]),
                                           precision=3, suppress_small=True))
        # # ------ adapt trait selector AFTER, with new traits:
        # log_trait_resp = sum(np.mean(g.item_logprob_by_tau(self.tau, item_index),
        #                              axis=0)
        #                      for g in groups.values())
        # # log_trait_resp[t] = mean logprob given t-th trait
        # self.trait_sel.adapt(log_trait_resp)
        # # **** self.trait_sel KL_div penalty must be calculated later, after TraitProb update ****
        logger.info(self.__class__.__name__ + f'[{self.item_index}].adapt(...) finished')
        return self

    def adapt_point(self, eta_0):
        """Find single MAP point estimate of eta,
        using same objective functions as sampler
        :param eta_0: 1D vector with start value
        :return: eta = 1D vector with new point estimate
            eta.shape == eta_0.shape == self.n_response_levels
        """
        def neg_logprob_1d(eta, *args):
            """Link to module-global function for 1D input
            :param eta: 1D vector with tentative parameter values
            :param args: tuple with additional arguments
            :return: nlp = scalar = - log p(data | eta)
            """
            return neg_logprob(eta[None, :], *args)[0]

        def d_neg_logprob_1d(eta, *args):
            """Gradient of neg_logprob_1d
            :param eta: 1D vector with tentative parameter values
            :param args: tuple with additional arguments
            :return: d_nlp = gradient vector
                d_nlp[i] = d neg_logprob_1d(eta,...) / d_eta[i]
                d_nlp.shape == eta.shape
            """
            return d_neg_logprob(eta[None, :], *args)[0]
        # --------------------------------------------------------
        res = minimize(neg_logprob_1d, eta_0,
                       jac=d_neg_logprob_1d,
                       args=self.sampler.args,
                       bounds=self.sampler.bounds,
                       tol=1e-4,
                       )
        if res.success:
            return res.x
        else:
            # ERROR exit
            logger.warning(res)  # ****
            return eta_0  # just let sampler do it anyway

    def relative_entropy_re_prior(self, trait_prob):
        """KL div(self || prior),
        estimated using current sampled approximation of self.eta
        AND KL div for current self.trait_sel
        :param trait_prob: single TraitProb object, prior for self.trait_sel
        :return: KL_div = scalar
        Method: E_{self}{ ln pdf(eta | self) / prior_pdf(eta)}
        """
        if self.n_samples == 1:
            h = 0.  # constant fix entropy
        else:
            h = entropy(self.eta)  # sample-estimated entropy
        return (self.trait_sel.relative_entropy_re_prior(trait_prob)
                - h - np.mean(prior_logprob(self.eta)))

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        Result: self.trait_sel pruned
        """
        self.trait_sel.prune(keep)

    def standardize(self, ss):
        """Standardize scale by down-scaling with given factor
        :param ss: 1D array with scale factors, one for each trait
        :return: None
        Method: all tau samples to be reduced by s toward zero
            logistic-transform to (0, 1) domain and adjust eta parameters there
        """
        s = np.dot(self.trait_sel.mean, ss)
        # = scalar factor for this item
        tau = self.tau / s
        n_samples = tau.shape[0]
        # tau[m, l] = m-th sample of new upper limit of l-th response interval
        p = logistic.cdf(tau)  # in (0, 1) domain
        p = np.concatenate((np.zeros((n_samples, 1)), p, np.ones((n_samples, 1))),
                           axis=1)
        w = np.diff(p, axis=1)
        # w[m, l] = width parameters in (0, 1) domain
        eta = np.log(w)
        self.latent_scale = 1. / s  # *********************
        # = new eta sample values defining rescaled tau values
        logger.debug(f'Scale[{self.item_index}]: Old mean tau = '
                     + np.array2string(np.mean(self.tau, axis=0),
                                       precision=3, suppress_small=True))
        self.sampler.x = eta
        # just to check for new values:
        logger.debug(f'Scale[{self.item_index}]: Rescaled mean tau = '
                     + np.array2string(np.mean(self.tau, axis=0),
                                       precision=3, suppress_small=True))


# -------------------------- internal module functions

def neg_logprob(eta, groups, item_index, trait_w):
    """Negative sum of log prob of observed data, given tentative samples.
    :param eta: 2D array with tentative sample values
        eta[n, l] = n-th sample of l-th log-width parameter
    :param groups: list with ref to all response groups
    :param item_index: integer index for this item
    :param trait_w: 1D weight vector
        trait_w[t] = weight for t-th trait
    :return: nlp = 1D array
        nlp[n] = negative log-prob for n-th tentative sample
        len(nlp) == eta.shape[0]
    """
    w = np.exp(eta)
    tau = _tau(w)
    lp = sum(g.item_logprob_by_tau(tau, item_index, trait_w)
             for g in groups)
    if np.any(np.isnan(lp)):
        logger.warning('Scale[{self.item_index}]: neg_logprob == nan. *** Should never happen!')
    return - lp - prior_logprob(eta)


def d_neg_logprob(eta, groups, item_index, trait_w):
    """Gradient of neg_logprob, given tentative samples.
    :param eta: 2D array with tentative sample values
        eta[n, l] = n-th sample of l-th log-width parameter
    :param groups: list with ref to all response groups
    :param item_index: integer index for this item
    :param trait_w: 1D weight vector
        trait_w[t] = weight for t-th trait
    :return: dnlp = 2D array
        dnlp[n, l] = d neg_logprob(eta[n]) / d eta[n, l]
        dnlp.shape == eta.shape
    """
    w = np.exp(eta)
    tau = _tau(w)
    if np.any(tau == -np.inf) or np.any(tau == np.inf):
        logger.warning('Scale[{self.item_index}]: Infinite tau. *** Should never happen!')
    d_lp_d_tau = sum(g.d_item_logprob_by_tau(tau, item_index, trait_w)
                     for g in groups)
    dlp = np.einsum('mk, mkl -> ml',
                    d_lp_d_tau, _d_tau_d_eta(w))
    return - dlp - d_prior_logprob(eta)


def prior_logprob(eta):
    """Prior log pdf of tentative eta array,
    disregarding the normalization constant,
    because the scale parameters are fixed.
    :param eta: 2D array of log interval-width parameters
        eta[n, l] = n-th sample of l-th log-width
    :return: lp = 1D array,
        lp[n] = log pdf(eta[n, :])
    """
    return - 0.5 * np.sum((eta / PRIOR_ETA_SCALE)**2, axis=-1)


def d_prior_logprob(eta):
    """Gradient of prior_logprob for tentative eta array,
    :param eta: 2D array of log interval-width parameters
        eta[n, l] = n-th sample of l-th log-width
    :return: dlp = 2D array,
        dlp[n, l] = d prior_logprob(eta[n,:] / d eta[n, l]
        dlp.shape == eta.shape
    """
    return - eta / PRIOR_ETA_SCALE**2


def _tau(w):
    """Response-interval internal limits, NOT including extremes at -+ inf
    :param w: 2D array of non-normalized interval-width parameters = np.exp(eta)
        w[n, l] = n-th sample of l-th interval non-normalized width in logistic-transformed domain
    :return: t = 2D array with corresponding bounds in the (-inf, inf) range,
        t.shape == (w.shape[0], w.shape[1] - 1)
    """
    cc = np.cumsum(w, axis=1)
    return logistic.ppf(cc[:, :-1] / cc[:, -1:])


def _d_tau_dw(w):
    """Jacobian of _tau(w)
    :param w: 2D array of non-normalized interval-width parameters = np.exp(eta)
        w[n, l] = proportional to n-th sample of l-th interval width
        in logistic-transformed domain
    :return: 3D array dt_dw, with
        dt_dw[n, l, j] = d _tau[n, l] / dw[n, j]
        dt_dw.shape == (w.shape[0], w.shape[1]-1, w.shape[1])
    Method:
    t_l = log(w_0 +...+ w_l) - log(w_{l+1} +...+ w_{L-1}); l = 0,..., L-2, any n
    dt_nlj / d w_nj =
        = +1 / (w_0 +...+ w_l); j <= l
        = -1 / (w_{l+1}+...+w_{L-1}); j > l
    2019-07-04 seems OK
    """
    nw = w.shape[1]
    cc = np.cumsum(w, axis=1)
    cc_plus = cc[:, :-1]
    cc_minus = cc[:, -1:] - cc_plus
    mask = np.ones((nw-1, nw))
    dt_dw = (1. / cc_plus[:, :, None] * np.tril(mask, k=0)
             - 1. / cc_minus[:, :, None] * np.triu(mask, k=1))
    return dt_dw


def _d_tau_d_eta(w):
    """Jacobian of _tau(w) w.r.t log(w)
    :param w: 2D array of non-normalized interval-width parameters = np.exp(eta)
        w[n, l] = proportional to n-th sample of l-th interval width
        in logistic-transformed domain
    :return: 3D array dt_d_eta with
        dt_d_eta[n, l, j] = d _tau[n, l] / d_eta[n, j]
        dt_d_eta.shape == (w.shape[0], w.shape[1]-1, w.shape[1])
    2019-07-04, sum(_d_tau_d_eta(w), axis=-1) == 0. as it should
    """
    return _d_tau_dw(w) * w[:, None, :]


def _cond_mean(tau, scale=1.):
    """Conditional means of trait values for each ordinal rating,
    for each rating-interval limits on the latent scale,
    assuming traits are normal-distributed with given scale
    :param tau: 2D array with upper interval limits for rating
        tau[m, l] = m-th sample of upper interval limit for rating == l
        lower limit is - inf for l == 0
        upper limit is + inf for l == n_rating_levels-1
        tau.shape[-1] == n_rating_levels - 1
    :return: theta = 2D array with mean traits
        theta[m, l] = m-th sample of conditional mean trait, given rating=l
        theta.shape == (tau.shape[0], n_rating_levels)
    """
    n_samples = tau.shape[0]
    n_rating_levels = tau.shape[1] + 1
    tau = np.concatenate((np.full((n_samples,1), - np.inf),
                          tau,
                          np.full((n_samples,1), np.inf)), axis=1)
    return np.array([[norm.expect(scale=scale, lb=tau_m[l], ub=tau_m[l+1],
                                  conditional=True)
                      for l in range(n_rating_levels)]
                     for tau_m in tau])


def _cond_median(tau, scale=1.):  # **** not used, but much faster than _cond_mean
    """Conditional medians of trait values for each ordinal rating,
    for each rating-interval limits on the latent scale,
    assuming traits are normal-distributed with given scale
    :param tau: 2D array with upper interval limits for rating
        tau[m, l] = m-th sample of upper interval limit for rating == l
        lower limit is - inf for l == 0
        upper limit is + inf for l == n_rating_levels-1
        tau.shape[-1] == n_rating_levels - 1
    :return: theta = 2D array with median trait values
        theta[m, l] = m-th sample of conditional median trait, given rating = l
        theta.shape == (tau.shape[0], n_rating_levels)
    """
    n_samples = tau.shape[0]
    p = norm.cdf(tau, scale=scale)
    p = np.concatenate((np.zeros((n_samples, 1)),
                        p,
                        np.ones((n_samples, 1))), axis=1)
    p_mid = (p[:, :-1] + p[:, 1:]) / 2
    return norm.ppf(p_mid, scale=scale)


# --------------------------------------------------- TEST:

if __name__ == '__main__':
    from .item_trait_selector import TraitProb

    # from scipy.optimize import check_grad, approx_fprime

    w = np.array([[1., 2.,3.,4.,5.]])
    w = np.concatenate((w, 2*w), axis=0)
    # 2 samples, 5 levels
    print('w =\n', w)
    tau = _tau(w)
    print('tau(w) = \n', tau)
    scale = 1.
    print(f'_cond_mean(tau, scale={scale}) = \n', _cond_mean(tau, scale=scale))
    print(f'_cond_median(tau, scale={scale}) = \n', _cond_median(tau, scale=scale))

    print('_d_tau_dw(w)= ', _d_tau_dw(w))

    dt_d_eta = _d_tau_d_eta(w)
    print('_d_tau_d_eta(w)= ', dt_d_eta)
    print('sum dt_d_eta= ', np.sum(dt_d_eta, axis=-1))

    trait_prob = TraitProb.initialize(5)
    item_sc = OrdinalItemScale.initialize(10 * w[0], n_samples=3, trait_scale=3.,
                                          trait_prob=trait_prob,
                                          item_index=0)
    print('item_sc= ', item_sc)
    print('item_sc.n_samples=', item_sc.n_samples)
    print('item_sc.n_response_levels=', item_sc.n_response_levels)
    print('item_sc.eta=\n', item_sc.eta)
    print('item_sc.tau=\n', item_sc.tau)
    # print('item_sc.typical_trait=\n', item_sc.typical_trait())
    r = [-1, 0, 1, 2, 3, 4]
    print('item_sc.sample_trait=\n', item_sc.sample_trait(r, 5))

    # ------- neg_loprob gradient tested in test_item_scale
